# -*- coding: utf-8 -*-

"""

LoggedFS-python
Filesystem monitoring with Fuse and Python
https://github.com/pleiszenburg/loggedfs-python

	src/loggedfs/_core/out.py: Log output formatting and filtering

	Copyright (C) 2017-2019 Sebastian M. Ernst <ernst@pleiszenburg.de>

<LICENSE_BLOCK>
The contents of this file are subject to the Apache License
Version 2 ("License"). You may not use this file except in
compliance with the License. You may obtain a copy of the License at
https://www.apache.org/licenses/LICENSE-2.0
https://github.com/pleiszenburg/loggedfs-python/blob/master/LICENSE

Software distributed under the License is distributed on an "AS IS" basis,
WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License for the
specific language governing rights and limitations under the License.
</LICENSE_BLOCK>

"""


# +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
# IMPORT
# +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

import base64
import errno
from functools import wraps
import grp
import inspect
import json
import pwd
import zlib

from fuse import (
	fuse_get_context,
	FuseOSError,
	)

from .ipc import send
from .log import log_msg
from .timing import time


# +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
# CONST
# +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

STATUS_DICT = {
	True: 'SUCCESS',
	False: 'FAILURE'
	}

ATTR_NAME = '__name__'
ATTR_ERRNO = 'errno'

NAME_OSERROR = 'OSError'
NAME_FUSEOSERROR = 'FuseOSError'
NAME_UNKNOWN = 'Unknown Exception'

ERROR_STAGE1 = 'UNEXPECTED in operation stage (1)'
ERROR_STAGE2 = 'UNEXPECTED in log stage (2)'


# +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
# ROUTINES
# +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

def event(format_pattern = ''):

	def wrapper(func):

		_func_param = inspect.signature(func).parameters
		func_arg_names = tuple(_func_param.keys())[1:]
		func_arg_abspath = func_arg_names[[item.endswith('path') for item in func_arg_names].index(True)]
		func_arg_defaults = {
			k: _func_param[k].default
			for k in func_arg_names
			if _func_param[k].default != inspect._empty
			}
		del _func_param

		@wraps(func)
		def wrapped(self, *func_args, **func_kwargs):

			ret_value = None
			ret_status = False
			try:
				ret_value = func(self, *func_args, **func_kwargs)
				ret_status = True
			except FuseOSError as e:
				ret_value = (NAME_FUSEOSERROR, e.errno)
				raise e
			except OSError as e:
				ret_value = (NAME_OSERROR, e.errno)
				raise FuseOSError(e.errno)
			except Exception as e:
				if hasattr(e, ATTR_ERRNO): # all subclasses of OSError
					ret_value = (getattr(type(e), ATTR_NAME, NAME_UNKNOWN), e.errno)
					raise FuseOSError(e.errno)
				else:
					self._logger.exception(log_msg(self._log_json, ERROR_STAGE1))
					raise e
			else:
				return ret_value
			finally:
				try:
					_log_event_(
						self,
						func, func_arg_names, func_arg_abspath, func_arg_defaults, func_args, func_kwargs,
						format_pattern,
						ret_status, ret_value
						)
				except Exception as e:
					self._logger.exception(log_msg(self._log_json, ERROR_STAGE2))
					raise e

		return wrapped

	return wrapper


def decode_buffer(in_buffer):

	if not isinstance(in_buffer, str):
		raise TypeError('in_buffer must be a string')

	return zlib.decompress(base64.b64decode(in_buffer.encode('utf-8')))


def _encode_buffer_(in_bytes):

	return base64.b64encode(zlib.compress(in_bytes, 1)).decode('utf-8') # compress level 1 (weak)


def _get_fh_from_fip_(fip):

	if fip is None:
		return -1
	if not hasattr(fip, 'fh'):
		return -2
	if not isinstance(fip.fh, int):
		return -3
	return fip.fh


def _get_group_name_from_gid_(gid):

	try:
		return grp.getgrgid(gid).gr_name
	except KeyError:
		return '[gid: omitted argument]'


def _get_process_cmdline_(pid):

	try:
		with open('/proc/%d/cmdline' % pid, 'r') as f: # TODO encoding, bytes?
			cmdline = f.read()
		return cmdline.replace('\x00', ' ').strip()
	except FileNotFoundError:
		return ''


def _get_user_name_from_uid_(uid):

	try:
		return pwd.getpwuid(uid).pw_name
	except KeyError:
		return '[uid: omitted argument]'


def _log_event_(
	self,
	func, func_arg_names, func_arg_abspath, func_arg_defaults, func_args, func_kwargs,
	format_pattern,
	ret_status, ret_value
	):

	uid, gid, pid = fuse_get_context()

	p_cmdname = ''
	if self._log_printprocessname or self._log_json:
		p_cmdname = _get_process_cmdline_(pid).strip()

	log_dict = {
		'proc_cmd': p_cmdname,
		'proc_uid': uid,
		'proc_uid_name': _get_user_name_from_uid_(uid),
		'proc_gid': gid,
		'proc_gid_name': _get_group_name_from_gid_(gid),
		'proc_pid': pid,
		'action': func.__name__,
		'status': ret_status,
		}

	arg_dict = {
		arg_name: arg
		for arg_name, arg in zip(func_arg_names, func_args)
		}
	arg_dict.update({
		arg_name: func_kwargs.get(arg_name, func_arg_defaults[arg_name])
		for arg_name in func_arg_names[len(func_args):]
		})

	try:
		arg_dict['uid_name'] = _get_user_name_from_uid_(arg_dict['uid'])
	except KeyError:
		pass
	try:
		arg_dict['gid_name'] = _get_group_name_from_gid_(arg_dict['gid'])
	except KeyError:
		pass
	try:
		arg_dict['fip'] = _get_fh_from_fip_(arg_dict['fip'])
	except KeyError:
		pass
	for k in arg_dict.keys():
		if k.endswith('path'):
			arg_dict[k] = self._full_path(arg_dict[k])
	try:
		arg_dict['buf_len'] = len(arg_dict['buf'])
		arg_dict['buf'] = _encode_buffer_(arg_dict['buf']) if self._log_buffers else ''
	except KeyError:
		pass

	log_dict.update({'param_%s' % k: v for k, v in arg_dict.items()})

	if log_dict['status']: # SUCCESS
		if any((
			ret_value is None,
			isinstance(ret_value, int),
			isinstance(ret_value, str),
			isinstance(ret_value, dict),
			isinstance(ret_value, list)
			)):
			log_dict['return'] = ret_value
		elif isinstance(ret_value, bytes):
			log_dict['return_len'] = len(ret_value)
			log_dict['return'] = _encode_buffer_(ret_value) if self._log_buffers else ''

	else: # FAILURE
		log_dict.update({
			'return': None,
			'return_exception': ret_value[0],
			'return_errno': ret_value[1],
			'return_errorcode': errno.errorcode[ret_value[1]]
			})

	if not self._lib_mode:
		if not self._log_filter.match(log_dict):
			return

	if self._log_json and not self._lib_mode:
		self._logger.info( json.dumps(log_dict, sort_keys = True)[1:-1] )
		return
	elif self._lib_mode:
		log_dict['time'] = time.time_ns()
		send(log_dict)
		return

	log_out = ' '.join([
		'%s %s' % (func.__name__, format_pattern.format(**log_dict)),
		'{%s}' % STATUS_DICT[ret_status],
		'[ pid = %d %suid = %d ]' % (pid, ('%s ' % p_cmdname) if len(p_cmdname) > 0 else '', uid),
		'( r = %s )' % str(log_dict['return'])
			if log_dict['status'] else
		'( %s = %d )' % ret_value
		])

	self._logger.info(log_out)



# -*- coding: utf-8 -*-

"""

LoggedFS-python
Filesystem monitoring with Fuse and Python
https://github.com/pleiszenburg/loggedfs-python

	src/loggedfs/_core/filter.py: Filtering events by criteria

	Copyright (C) 2017-2019 Sebastian M. Ernst <ernst@pleiszenburg.de>

<LICENSE_BLOCK>
The contents of this file are subject to the Apache License
Version 2 ("License"). You may not use this file except in
compliance with the License. You may obtain a copy of the License at
https://www.apache.org/licenses/LICENSE-2.0
https://github.com/pleiszenburg/loggedfs-python/blob/master/LICENSE

Software distributed under the License is distributed on an "AS IS" basis,
WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License for the
specific language governing rights and limitations under the License.
</LICENSE_BLOCK>

"""


# +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
# IMPORT
# +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

from collections import OrderedDict
import re

import xmltodict

from .defaults import LOG_ENABLED_DEFAULT, LOG_PRINTPROCESSNAME_DEFAULT


# +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
# FILTER FIELD CLASS
# +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

class filter_field_class:


	def __init__(self, name, value):

		self._name_is_func = hasattr(name, '__call__')

		if not self._name_is_func and not isinstance(name, str):
			raise TypeError('name must either be callable or a string')
		if not hasattr(value, '__call__'):
			raise TypeError('value must either be callable')

		self._name = name
		self._value = value


	def __repr__(self):

		return '<filter_field name="%s" value="%s"/>' % (
			self._name if not self._name_is_func else '{callable:' + getattr(self._name, '__name__', 'NONAME') + '}',
			'{callable:' + getattr(self._value, '__name__', 'NONAME') + '}'
			)


	@property
	def name_is_func(self):

		return self._name_is_func


	@property
	def name(self):

		return self._name


	@property
	def value(self):

		return self._value


# +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
# FILTER ITEM CLASS
# +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

class filter_item_class:


	def __init__(self, fields_list):

		if not isinstance(fields_list, list):
			raise TypeError('fields_list must be of type list')
		if len(fields_list) == 0:
			raise ValueError('at least one field is required for setting up a filter')
		if any((not isinstance(item, filter_field_class) for item in fields_list)):
			raise TypeError('fields_list must only contain type filter_field_class')

		self._fields_list = fields_list
		self._field_names = {field.name for field in self._fields_list if not field.name_is_func}
		self._field_nofuncs = {field for field in self._fields_list if not field.name_is_func}
		self._field_funcs = {field for field in self._fields_list if field.name_is_func}


	def __repr__(self):

		return (
			'<filter_item>\n\t' +
			'\n\t'.join((repr(field) for field in self._fields_list))
			+ '\n</filter_item>'
			)


	def match(self, event_dict):

		if not isinstance(event_dict, dict):
			raise TypeError('event_dict must be of type dict')

		if len(self._field_names - event_dict.keys()) > 0:
			return False

		if any((not field.value(event_dict[field.name]) for field in self._field_nofuncs)):
			return False

		for field in self._field_funcs:
			key_match = None
			for key in event_dict.keys():
				if field.name(key):
					key_match = key
					break
			if key_match is None:
				return False
			if not field.value(event_dict[key_match]):
				return False

		return True


	@staticmethod
	def _from_xmldict(xml_dict):

		if not isinstance(xml_dict, OrderedDict) and not isinstance(xml_dict, dict):
			raise TypeError('can not construct filter item from non-dict type')
		if any((not isinstance(item, str) for item in xml_dict.keys())):
			raise TypeError('non-string key in dict')
		if any((not isinstance(item, str) for item in xml_dict.values())):
			raise TypeError('non-string value in dict')

		fields_list = []

		try:
			if xml_dict['@retname'] == 'SUCCESS':
				fields_list.append(filter_field_class('status', lambda x: x == True))
			elif xml_dict['@retname'] == 'FAILURE':
				fields_list.append(filter_field_class('status', lambda x: x == False))
			elif xml_dict['@retname'] != '.*':
				raise ValueError('unexpected value for "retname"')
		except KeyError:
			pass

		try:
			if xml_dict['@extension'] != '.*':
				fields_list.append(filter_field_class(
					lambda x: x.endswith('path'), re.compile(xml_dict['@extension']).match
					))
		except KeyError:
			pass

		try:
			if xml_dict['@uid'].isdecimal():
				fields_list.append(filter_field_class(
					'proc_uid', lambda x: x == int(xml_dict['@uid'])
					))
			elif xml_dict['@uid'] != '*':
				raise ValueError('unexpected value for "uid"')
		except KeyError:
			pass

		try:
			if xml_dict['@action'] != '.*':
				fields_list.append(filter_field_class(
					'action', re.compile(xml_dict['@action']).match
					))
		except KeyError:
			pass

		try:
			if xml_dict['@command'] != '.*':
				fields_list.append(filter_field_class(
					'proc_cmd', re.compile(xml_dict['@command']).match
					))
		except KeyError:
			pass

		if len(fields_list) > 0:
			return filter_item_class(fields_list)
		else:
			return None


# +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
# FILTER PIPELINE CLASS
# +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

class filter_pipeline_class:


	VALID_XML_BLOCKS = ('@logEnabled', '@printProcessName', 'includes', 'excludes')


	def __init__(self, include_list = None, exclude_list = None):

		if include_list is None:
			include_list = []
		if exclude_list is None:
			exclude_list = []

		if not isinstance(include_list, list):
			raise TypeError('include_list must have type list')
		if not isinstance(exclude_list, list):
			raise TypeError('exclude_list must have type list')
		if any((not isinstance(item, filter_item_class) for item in include_list)):
			raise TypeError('include_list must only contain type filter_item_class')
		if any((not isinstance(item, filter_item_class) for item in exclude_list)):
			raise TypeError('exclude_list must only contain type filter_item_class')

		self._include_list = include_list
		self._exclude_list = exclude_list


	def __repr__(self):

		return (
			'<filter_pipeline>\n'
			+ '\t<include>\n'
			+ '\n'.join( (
				'\n'.join(('\t\t' + line for line in repr(item).split('\n') ))
				for item in self._include_list)
				) + ('\n' if len(self._include_list) > 0 else '')
			+ '\t</include>\n'
			+ '\t<exclude>\n'
			+ '\n'.join( (
				'\n'.join(('\t\t' + line for line in repr(item).split('\n') ))
				for item in self._exclude_list)
				) + ('\n' if len(self._exclude_list) > 0 else '')
			+ '\t</exclude>\n'
			+ '<filter_pipeline>'
			)


	def match(self, event_dict):

		if not isinstance(event_dict, dict):
			raise TypeError('event_dict must be of type dict')

		if len(self._include_list) > 0:
			if not any((item.match(event_dict) for item in self._include_list)):
				return False

		if any((item.match(event_dict) for item in self._exclude_list)):
			return False

		return True


	@staticmethod
	def from_xmlstring(xml_str):
		"""Parse XML configuration string and return instance of filter_pipeline_class.
		Compatibility layer for original LoggedFS XML configuration file format.
		"""

		if not isinstance(xml_str, str):
			raise TypeError('xml_str must have type str')
		if len(xml_str) == 0:
			raise ValueError('xml_str must not be empty')
		try:
			xml_dict = xmltodict.parse(xml_str)
		except:
			raise ValueError('xml_str does not contain valid XML')
		try:
			xml_dict = xml_dict['loggedFS']
		except KeyError:
			raise ValueError('XML tree does not have loggedFS top-level tag')
		if len(xml_dict.keys() - set(filter_pipeline_class.VALID_XML_BLOCKS)) > 0:
			raise ValueError('unexpected tags and/or parameters in XML tree')

		log_enabled = xml_dict.pop('@logEnabled', None)
		if log_enabled is None:
			log_enabled = LOG_ENABLED_DEFAULT
		else:
			log_enabled = log_enabled.lower() == 'true'
		log_printprocessname = xml_dict.pop('@printProcessName', None)
		if log_printprocessname is None:
			log_printprocessname = LOG_PRINTPROCESSNAME_DEFAULT
		else:
			log_printprocessname = log_printprocessname.lower() == 'true'

		group_list = []
		for f_type in ('includes', 'excludes'):
			group = xml_dict.pop(f_type, None)
			if group is None:
				group_list.append([])
				continue
			if not isinstance(group, OrderedDict) and not isinstance(group, dict):
				raise TypeError('malformed XML tree for %s' % f_type)
			group = group.get(f_type[:-1], None)
			if group is None:
				group_list.append([])
				continue
			if not any((
				isinstance(group, list), isinstance(group, OrderedDict), isinstance(group, dict)
				)):
				raise TypeError('malformed XML tree for %s' % f_type[:-1])
			if isinstance(group, list):
				tmp = [filter_item_class._from_xmldict(item) for item in group]
				tmp = [item for item in tmp if item is not None]
				group_list.append(tmp)
			else:
				tmp = filter_item_class._from_xmldict(group)
				if tmp is not None:
					group_list.append([tmp])
				else:
					group_list.append([])

		return log_enabled, log_printprocessname, filter_pipeline_class(*group_list)



# -*- coding: utf-8 -*-

"""

LoggedFS-python
Filesystem monitoring with Fuse and Python
https://github.com/pleiszenburg/loggedfs-python

	src/loggedfs/_core/notify.py: Notification backend - LoggedFS as a library

	Copyright (C) 2017-2019 Sebastian M. Ernst <ernst@pleiszenburg.de>

<LICENSE_BLOCK>
The contents of this file are subject to the Apache License
Version 2 ("License"). You may not use this file except in
compliance with the License. You may obtain a copy of the License at
https://www.apache.org/licenses/LICENSE-2.0
https://github.com/pleiszenburg/loggedfs-python/blob/master/LICENSE

Software distributed under the License is distributed on an "AS IS" basis,
WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License for the
specific language governing rights and limitations under the License.
</LICENSE_BLOCK>

"""


# +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
# IMPORT
# +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

import atexit
import inspect
import os
import subprocess
import sys
import threading

from .defaults import (
	FUSE_ALLOWOTHER_DEFAULT,
	LOG_BUFFERS_DEFAULT
	)
from .filter import filter_pipeline_class
from .ipc import receive, end_of_transmission


# +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
# CONST
# +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

# https://en.wikipedia.org/wiki/ANSI_escape_code
c = {
	'RESET': '\033[0;0m',
	'BOLD': '\033[;1m',
	'REVERSE': '\033[;7m',
	'GREY': '\033[1;30m',
	'RED': '\033[1;31m',
	'GREEN': '\033[1;32m',
	'YELLOW': '\033[1;33m',
	'BLUE': '\033[1;34m',
	'MAGENTA': '\033[1;35m',
	'CYAN': '\033[1;36m',
	'WHITE': '\033[1;37m'
	}


# +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
# CLASS: NOTIFY
# +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

class notify_class:
	"""Simple wrapper for using LoggedFS-python as a library.
	Attach a method of your choice to filesystem events.
	"""


	def __init__(self,
		directory,
		consumer_out_func = None, # consumes signals
		consumer_err_func = None, # consumes anything on stderr
		post_exit_func = None, # called on exit
		log_filter = None,
		log_buffers = LOG_BUFFERS_DEFAULT,
		fuse_allowother = FUSE_ALLOWOTHER_DEFAULT,
		background = False # thread in background
		):
		"""Creates a filesystem notifier object.

		- directory: Relative or absolute path as a string
		- consumer_out_func: None or callable, consumes events provided as a dictionary
		- consumer_err_func: None or callable, consumes output from stderr
		- post_exit_func: None or callable, called when notifier was terminated
		- log_filter: None or instance of filter_pipeline_class
		- log_buffers: Boolean, activates logging of read and write buffers
		- fuse_allowother: Boolean, allows other users to see the LoggedFS filesystem
		- background: Boolean, starts notifier in a thread
		"""

		if log_filter is None:
			log_filter = filter_pipeline_class()

		if not isinstance(directory, str):
			raise TypeError('directory must be of type string')
		if not os.path.isdir(directory):
			raise ValueError('directory must be a path to an existing directory')
		if not os.access(directory, os.W_OK | os.R_OK):
			raise ValueError('not sufficient permissions on directory')

		if consumer_out_func is not None and not hasattr(consumer_out_func, '__call__'):
			raise TypeError('consumer_out_func must either be None or callable')
		if hasattr(consumer_out_func, '__call__'):
			if len(inspect.signature(consumer_out_func).parameters.keys()) != 1:
				raise ValueError('consumer_out_func must have one parameter')
		if consumer_err_func is not None and not hasattr(consumer_err_func, '__call__'):
			raise TypeError('consumer_err_func must either be None or callable')
		if hasattr(consumer_err_func, '__call__'):
			if len(inspect.signature(consumer_err_func).parameters.keys()) != 1:
				raise ValueError('consumer_err_func must have one parameter')
		if post_exit_func is not None and not hasattr(post_exit_func, '__call__'):
			raise TypeError('post_exit_func must either be None or callable')
		if not isinstance(log_filter, filter_pipeline_class):
			raise TypeError('log_filter must either be None or of type filter_pipeline_class')
		if not isinstance(log_buffers, bool):
			raise TypeError('log_buffers must be of type bool')
		if not isinstance(fuse_allowother, bool):
			raise TypeError('fuse_allowother must be of type bool')
		if not isinstance(background, bool):
			raise TypeError('background must be of type bool')

		self._directory = os.path.abspath(directory)
		self._post_exit_func = post_exit_func
		self._consumer_out_func = consumer_out_func
		self._consumer_err_func = consumer_err_func
		self._log_filter = log_filter
		self._log_buffers = log_buffers
		self._fuse_allowother = fuse_allowother
		self._background = background

		self._up = True

		command = ['loggedfs',
			'-f', # foreground
			'-s', # no syslog
			'--lib' # "hidden" library mode
			]
		if self._log_buffers:
			command.append('-b') # also log read and write buffers
		if self._fuse_allowother:
			command.append('-p')
		command.append(self._directory)

		args = (command, self._handle_stdout, self._handle_stderr, self._handle_exit)

		atexit.register(self.terminate)
		if self._background:
			self._t = threading.Thread(target = receive, args = args, daemon = False)
			self._t.start()
		else:
			receive(*args)


	def _handle_stderr(self, msg):

		if self._consumer_err_func is not None:
			self._consumer_err_func(msg)
		else:
			sys.stderr.write(c['RED'] + str(msg).rstrip('\n') + c['RESET'] + '\n')
			sys.stderr.flush()


	def _handle_stdout(self, msg):

		if not isinstance(msg, end_of_transmission):
			if not self._log_filter.match(msg):
				return

		if self._consumer_out_func is not None:
			self._consumer_out_func(msg)
		else:
			sys.stdout.write(c['GREEN'] + str(msg) + c['RESET'] + '\n')
			sys.stdout.flush()


	def _handle_exit(self):

		if self._post_exit_func is not None:
			self._post_exit_func()


	def terminate(self):

		if not self._up:
			return

		self._up = False

		command = ['fusermount', '-u', self._directory]
		proc = subprocess.Popen(command, stdout = subprocess.PIPE, stderr = subprocess.PIPE)
		proc.wait()

		if not self._background:
			self._t.join()



# -*- coding: utf-8 -*-

"""

LoggedFS-python
Filesystem monitoring with Fuse and Python
https://github.com/pleiszenburg/loggedfs-python

	src/loggedfs/_core/ipc.py: Fast IPC through a pipe

	Copyright (C) 2017-2019 Sebastian M. Ernst <ernst@pleiszenburg.de>

<LICENSE_BLOCK>
The contents of this file are subject to the Apache License
Version 2 ("License"). You may not use this file except in
compliance with the License. You may obtain a copy of the License at
https://www.apache.org/licenses/LICENSE-2.0
https://github.com/pleiszenburg/loggedfs-python/blob/master/LICENSE

Software distributed under the License is distributed on an "AS IS" basis,
WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License for the
specific language governing rights and limitations under the License.
</LICENSE_BLOCK>

"""


# +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
# IMPORT
# +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

import pickle
import queue
import struct
import subprocess
import sys
import threading
import time


# +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
# CONST
# +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

PREFIX = b'\xBA\xDE\xAF\xFE'
LEN_DTYPE = 'Q' # uint64
WAIT_TIMEOUT = 0.1 # seconds


# +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
# CLASS: END OF TRANSMISSION
# +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

class end_of_transmission:


	def __init__(self, id):

		self._id = id


	@property
	def id(self):
		return self._id


	def __repr__(self):

		return '<end of transmission on stream "{ID}">'.format(ID = self._id)


# +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
# CLASS: RECEIVER (SINGLE STREAM)
# +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

class _receiver_class:


	def __init__(self, stream_id, in_stream, decoder_func, processing_func):

		self._id = stream_id
		self._s = in_stream
		self._f = processing_func
		self._q = queue.Queue()
		self._t = threading.Thread(
			target = decoder_func,
			args = (self._id, self._s, self._q),
			daemon = True
			)
		self._t.start()
		self.join = self._t.join


	def flush(self):

		while not self._q.empty():
			try:
				data = self._q.get_nowait()
			except queue.Empty:
				pass
			else:
				self._f(data)
				self._q.task_done()


# +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
# ROUTINES: DECODER FUNCTIONS FOR RECEIVER
# +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

def _out_decoder(_id, _s, _q):

	prefix_len = len(PREFIX)
	while True:
		prefix = _s.read(prefix_len)
		if len(prefix) == 0: # end of pipe
			_q.put(end_of_transmission(_id))
			break
		data_len_encoded = _s.read(8)
		data_len = struct.unpack(LEN_DTYPE, data_len_encoded)[0]
		data_bin = _s.read(data_len)
		_q.put(pickle.loads(data_bin))


def _err_decoder(_id, _s, _q):

	while True:
		msg = _s.readline()
		if len(msg) == 0:
			_q.put(end_of_transmission(_id))
			break
		_q.put(msg.decode('utf-8'))


# +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
# ROUTINES: SEND AND RECEIVE
# +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

def receive(cmd_list, out_func, err_func, post_exit_func):

		proc = subprocess.Popen(cmd_list, stdout = subprocess.PIPE, stderr = subprocess.PIPE)
		proc_alive = True
		out_r = _receiver_class('out', proc.stdout, _out_decoder, out_func)
		err_r = _receiver_class('err', proc.stderr, _err_decoder, err_func)

		while proc_alive:
			time.sleep(WAIT_TIMEOUT)
			out_r.flush()
			err_r.flush()
			proc_alive = proc.poll() is None

		out_r.join()
		err_r.join()
		post_exit_func()


def send(data):

	data_bin = pickle.dumps(data)
	data_len = struct.pack(LEN_DTYPE, len(data_bin))
	sys.stdout.buffer.write(PREFIX)
	sys.stdout.buffer.write(data_len)
	sys.stdout.buffer.write(data_bin)
	sys.stdout.flush()



# -*- coding: utf-8 -*-

"""

LoggedFS-python
Filesystem monitoring with Fuse and Python
https://github.com/pleiszenburg/loggedfs-python

	src/loggedfs/_core/fs.py: File system core

	Copyright (C) 2017-2019 Sebastian M. Ernst <ernst@pleiszenburg.de>

<LICENSE_BLOCK>
The contents of this file are subject to the Apache License
Version 2 ("License"). You may not use this file except in
compliance with the License. You may obtain a copy of the License at
https://www.apache.org/licenses/LICENSE-2.0
https://github.com/pleiszenburg/loggedfs-python/blob/master/LICENSE

Software distributed under the License is distributed on an "AS IS" basis,
WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License for the
specific language governing rights and limitations under the License.
</LICENSE_BLOCK>

"""


# +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
# IMPORT
# +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

import errno
import os
import stat

from fuse import (
	FUSE,
	fuse_get_context,
	FuseOSError,
	Operations,
	UTIME_NOW,
	UTIME_OMIT,
	)
try:
	from fuse import __features__ as fuse_features
except ImportError:
	fuse_features = {}

from .defaults import (
	FUSE_ALLOWOTHER_DEFAULT,
	FUSE_FOREGROUND_DEFAULT,
	LIB_MODE_DEFAULT,
	LOG_BUFFERS_DEFAULT,
	LOG_ENABLED_DEFAULT,
	LOG_JSON_DEFAULT,
	LOG_PRINTPROCESSNAME_DEFAULT,
	LOG_SYSLOG_DEFAULT
	)
from .filter import filter_pipeline_class
from .log import get_logger, log_msg
from .out import event
from .timing import time


# +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
# ROUTINES
# +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

def loggedfs_factory(directory, **kwargs):

	if not isinstance(directory, str):
		raise TypeError('directory must be of type string')
	if not os.path.isdir(directory):
		raise ValueError('directory must be a path to an existing directory')

	if not isinstance(kwargs.get('fuse_foreground', FUSE_FOREGROUND_DEFAULT), bool):
		raise TypeError('fuse_foreground must be of type bool')
	if not isinstance(kwargs.get('fuse_allowother', FUSE_ALLOWOTHER_DEFAULT), bool):
		raise TypeError('fuse_allowother must be of type bool')

	return FUSE(
		_loggedfs(
			directory,
			**kwargs
			),
		directory,
		raw_fi = True,
		nothreads = True,
		foreground = kwargs.get('fuse_foreground', FUSE_FOREGROUND_DEFAULT),
		allow_other = kwargs.get('fuse_allowother', FUSE_ALLOWOTHER_DEFAULT),
		default_permissions = kwargs.get('fuse_allowother', FUSE_ALLOWOTHER_DEFAULT),
		attr_timeout = 0,
		entry_timeout = 0,
		negative_timeout = 0,
		sync_read = False, # relying on fuse.Operations class defaults?
		# max_readahead = 0, # relying on fuse.Operations class defaults?
		# direct_io = True, # relying on fuse.Operations class defaults?
		nonempty = True, # common options taken from LoggedFS
		use_ino = True # common options taken from LoggedFS
		)


# +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
# CORE CLASS: Init and internal routines
# +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

class _loggedfs(Operations):


	flag_utime_omit_ok = 1


	requested_features = {
		'nanosecond_int': True,
		'utime_omit_none': True,
		'utime_now_auto': True
		}


	_ST_FIELDS = tuple(i for i in dir(os.stat_result) if i.startswith('st_'))
	_STVFS_FIELDS = tuple(i for i in dir(os.statvfs_result) if i.startswith('f_'))


	def __init__(self,
		directory,
		fuse_foreground = FUSE_FOREGROUND_DEFAULT,
		fuse_allowother = FUSE_ALLOWOTHER_DEFAULT,
		lib_mode = LIB_MODE_DEFAULT,
		log_buffers = LOG_BUFFERS_DEFAULT,
		log_enabled = LOG_ENABLED_DEFAULT,
		log_file = None,
		log_filter = None,
		log_json = LOG_JSON_DEFAULT,
		log_printprocessname = LOG_PRINTPROCESSNAME_DEFAULT,
		log_syslog = LOG_SYSLOG_DEFAULT,
		**kwargs
		):

		if log_filter is None:
			log_filter = filter_pipeline_class()

		if not isinstance(directory, str):
			raise TypeError('directory must be of type string')
		if not os.path.isdir(directory):
			raise ValueError('directory must be a path to an existing directory')
		if not os.access(directory, os.W_OK | os.R_OK):
			raise ValueError('not sufficient permissions on "directory"')

		if not isinstance(log_filter, filter_pipeline_class):
			raise TypeError('log_filter must either be None or of type filter_pipeline_class')
		if log_file is not None:
			if not os.path.isdir(os.path.dirname(log_file)):
				raise ValueError('path to logfile directory does not exist')
			if os.path.exists(log_file) and not os.path.isfile(log_file):
				raise ValueError('logfile exists and is not a file')
			if os.path.isfile(log_file) and not os.access(log_file, os.W_OK):
				raise ValueError('logfile exists and is not writeable')
			if not os.path.exists(log_file) and not os.access(directory, os.W_OK):
				raise ValueError('path to logfile directory is not writeable')
		if not isinstance(log_syslog, bool):
			raise TypeError('log_syslog must be of type bool')
		if not isinstance(log_enabled, bool):
			raise TypeError('log_enabled must be of type bool')
		if not isinstance(log_printprocessname, bool):
			raise TypeError('log_printprocessname must be of type bool')
		if not isinstance(log_json, bool):
			raise TypeError('log_json must be of type bool')
		if not isinstance(log_buffers, bool):
			raise TypeError('log_buffers must be of type bool')
		if not isinstance(lib_mode, bool):
			raise TypeError('lib_mode must be of type bool')

		if not isinstance(fuse_foreground, bool):
			raise TypeError('fuse_foreground must be of type bool')
		if not isinstance(fuse_allowother, bool):
			raise TypeError('fuse_allowother must be of type bool')

		self._root_path = directory
		self._log_printprocessname = log_printprocessname
		self._log_json = log_json
		self._log_buffers = log_buffers
		self._log_filter = log_filter
		self._lib_mode = lib_mode

		self._logger = get_logger('LoggedFS-python', log_enabled, log_file, log_syslog, self._log_json)

		if fuse_foreground:
			self._logger.info(log_msg(self._log_json, 'LoggedFS-python not running as a daemon'))
		if fuse_allowother:
			self._logger.info(log_msg(self._log_json, 'LoggedFS-python running as a public filesystem'))
		if log_file is not None:
			self._logger.info(log_msg(self._log_json, 'LoggedFS-python log file: %s' % log_file))

		self._logger.info(log_msg(self._log_json, 'LoggedFS-python starting at %s' % directory))

		try:
			self._root_path_fd = os.open(directory, os.O_RDONLY)
		except Exception as e:
			self._logger.exception('Directory access failed.')
			raise e

		log_configfile = kwargs.pop('_log_configfile', None)
		if log_configfile is not None:
			self._logger.info(log_msg(self._log_json,
				'LoggedFS-python using configuration file %s' % log_configfile
				))

		for flag_name in self.requested_features.keys():
			setattr(
				self,
				'flag_' + flag_name,
				self.requested_features[flag_name] and fuse_features.get(flag_name, False)
				)

		if len(kwargs) > 0:
			raise ValueError('unknown keyword argument(s)')


	def _full_path(self, partial_path):

		if partial_path.startswith('/'):
			partial_path = partial_path[1:]
		path = os.path.join(self._root_path, partial_path)
		return path


	@staticmethod
	def _rel_path(partial_path):

		if len(partial_path) == 0:
			return '.'
		elif partial_path == '/':
			return '.'
		elif partial_path.startswith('/'):
			return partial_path[1:]
		elif partial_path.startswith('./'):
			return partial_path[2:]
		else:
			return partial_path


# +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
# CORE CLASS: Filesystem & file methods - STUBS
#  ... addressing https://github.com/fusepy/fusepy/issues/81
# +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

	def create(self, path, mode, fi = None):

		raise FuseOSError(errno.ENOSYS)


	def flush(self, path, fip):

		raise FuseOSError(errno.ENOSYS)


	def fsync(self, path, datasync, fip):

		raise FuseOSError(errno.ENOSYS) # the original loggedfs just returns 0


	def ioctl(self, path, cmd, arg, fh, flags, data):

		raise FuseOSError(errno.ENOSYS)


	def lock(self, path, fh, cmd, lock):

		raise FuseOSError(errno.ENOSYS)


# +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
# CORE CLASS: Filesystem & file methods - IMPLEMENTATION
# +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

	@event(format_pattern = '{param_path}')
	def access(self, path, mode):

		if not os.access(self._rel_path(path), mode, dir_fd = self._root_path_fd):
			raise FuseOSError(errno.EACCES)


	@event(format_pattern = '{param_path} to {param_mode}')
	def chmod(self, path, mode):

		os.chmod(self._rel_path(path), mode, dir_fd = self._root_path_fd)


	@event(format_pattern = '{param_path} to {param_uid_name}({param_uid}):{param_gid_name}({param_gid})')
	def chown(self, path, uid, gid):

		os.chown(self._rel_path(path), uid, gid, dir_fd = self._root_path_fd, follow_symlinks = False)


	@event(format_pattern = '{param_path}')
	def destroy(self, path):

		os.close(self._root_path_fd)


	@event(format_pattern = '{param_path} (fh={param_fip})')
	def getattr(self, path, fip):

		if not fip:
			try:
				st = os.lstat(self._rel_path(path), dir_fd = self._root_path_fd)
			except FileNotFoundError:
				raise FuseOSError(errno.ENOENT)
		else:
			st = os.fstat(fip.fh)

		ret_dict = {key: getattr(st, key) for key in self._ST_FIELDS}

		for key in ['st_atime', 'st_ctime', 'st_mtime']:
			if self.flag_nanosecond_int:
				ret_dict[key] = ret_dict.pop(key + '_ns')
			else:
				ret_dict.pop(key + '_ns')

		return ret_dict


	@event(format_pattern = '{param_path}')
	def init(self, path):

		pass


	@event(format_pattern = '{param_source_path} to {param_target_path}')
	def link(self, target_path, source_path):

		target_rel_path = self._rel_path(target_path)

		os.link(
			self._rel_path(source_path), target_rel_path,
			src_dir_fd = self._root_path_fd, dst_dir_fd = self._root_path_fd
			)

		uid, gid, pid = fuse_get_context()
		os.chown(target_rel_path, uid, gid, dir_fd = self._root_path_fd, follow_symlinks = False)


	@event(format_pattern = '{param_path} {param_mode}')
	def mkdir(self, path, mode):

		rel_path = self._rel_path(path)

		os.mkdir(rel_path, mode, dir_fd = self._root_path_fd)

		uid, gid, pid = fuse_get_context()

		os.chown(rel_path, uid, gid, dir_fd = self._root_path_fd, follow_symlinks = False)
		os.chmod(rel_path, mode, dir_fd = self._root_path_fd) # follow_symlinks = False


	@event(format_pattern = '{param_path} {param_mode}')
	def mknod(self, path, mode, dev):

		rel_path = self._rel_path(path)

		if stat.S_ISREG(mode):
			res = os.open(
				rel_path, os.O_CREAT | os.O_EXCL | os.O_WRONLY, mode,
				dir_fd = self._root_path_fd
				) # TODO broken, applies umask to mode no matter what ...
			if res >= 0:
				os.close(res)
		elif stat.S_ISFIFO(mode):
			os.mkfifo(rel_path, mode, dir_fd = self._root_path_fd)
		else:
			os.mknod(rel_path, mode, dev, dir_fd = self._root_path_fd)

		uid, gid, pid = fuse_get_context()
		os.chown(rel_path, uid, gid, dir_fd = self._root_path_fd, follow_symlinks = False)
		os.chmod(rel_path, mode, dir_fd = self._root_path_fd) # follow_symlinks = False


	@event(format_pattern = '({param_fip}) {param_path} (fh={param_fip})')
	def open(self, path, fip):

		fip.fh = os.open(self._rel_path(path), fip.flags, dir_fd = self._root_path_fd)

		return 0


	@event(format_pattern = '{param_length} bytes from {param_path} at offset {param_offset} (fh={param_fip})')
	def read(self, path, length, offset, fip):

		ret = os.pread(fip.fh, length, offset)

		return ret


	@event(format_pattern = '{param_path}')
	def readdir(self, path, fh):

		rel_path = self._rel_path(path)

		dirents = ['.', '..']
		if stat.S_ISDIR(os.lstat(rel_path, dir_fd = self._root_path_fd).st_mode):
			dir_fd = os.open(rel_path, os.O_RDONLY, dir_fd = self._root_path_fd)
			dirents.extend(os.listdir(dir_fd))
			os.close(dir_fd)

		return dirents


	@event(format_pattern = '{param_path}')
	def readlink(self, path):

		pathname = os.readlink(self._rel_path(path), dir_fd = self._root_path_fd)

		if pathname.startswith('/'): # TODO check this ... actually required?
			return os.path.relpath(pathname, self._root_path)
		else:
			return pathname


	@event(format_pattern = '{param_path} (fh={param_fip})')
	def release(self, path, fip):

		os.close(fip.fh)


	@event(format_pattern = '{param_old_path} to {param_new_path}')
	def rename(self, old_path, new_path):

		os.rename(
			self._rel_path(old_path), self._rel_path(new_path),
			src_dir_fd = self._root_path_fd, dst_dir_fd = self._root_path_fd
			)


	@event(format_pattern = '{param_path}')
	def rmdir(self, path):

		os.rmdir(self._rel_path(path), dir_fd = self._root_path_fd)


	@event(format_pattern = '{param_path}')
	def statfs(self, path):

		fd = os.open(self._rel_path(path), os.O_RDONLY, dir_fd = self._root_path_fd)
		stv = os.statvfs(fd)
		os.close(fd)

		return {key: getattr(stv, key) for key in self._STVFS_FIELDS}


	@event(format_pattern = 'from {param_source_path} to {param_target_path_}')
	def symlink(self, target_path_, source_path):

		target_rel_path = self._rel_path(target_path_)

		os.symlink(source_path, target_rel_path, dir_fd = self._root_path_fd)

		uid, gid, pid = fuse_get_context()
		os.chown(target_rel_path, uid, gid, dir_fd = self._root_path_fd, follow_symlinks = False)


	@event(format_pattern = '{param_path} to {param_length} bytes (fh={param_fip})')
	def truncate(self, path, length, fip = None):

		if fip is None:

			fd = os.open(self._rel_path(path), flags = os.O_WRONLY, dir_fd = self._root_path_fd)
			ret = os.ftruncate(fd, length)
			os.close(fd)
			return ret

		else:

			return os.ftruncate(fip.fh, length)


	@event(format_pattern = '{param_path}')
	def unlink(self, path):

		os.unlink(self._rel_path(path), dir_fd = self._root_path_fd)


	@event(format_pattern = '{param_path}')
	def utimens(self, path, times = None):

		def _fix_time_(atime, mtime):
			if any(val in (atime, mtime) for val in [UTIME_OMIT, None]):
				st = os.lstat(relpath, dir_fd = self._root_path_fd)
				if atime in [UTIME_OMIT, None]:
					atime = st.st_atime_ns
				if mtime in [UTIME_OMIT, None]:
					mtime = st.st_mtime_ns
			if UTIME_NOW in (atime, mtime):
				now = time.time_ns()
				if atime == UTIME_NOW:
					atime = now
				if mtime == UTIME_NOW:
					mtime = now
			return (atime, mtime)

		relpath = self._rel_path(path)

		if self.flag_nanosecond_int:
			os.utime(relpath, ns = _fix_time_(*times), dir_fd = self._root_path_fd, follow_symlinks = False)
		else:
			os.utime(relpath, times = times, dir_fd = self._root_path_fd, follow_symlinks = False)


	@event(format_pattern = '{param_buf_len} bytes to {param_path} at offset {param_offset} (fh={param_fip})')
	def write(self, path, buf, offset, fip):

		res = os.pwrite(fip.fh, buf, offset)

		return res



# -*- coding: utf-8 -*-

"""

LoggedFS-python
Filesystem monitoring with Fuse and Python
https://github.com/pleiszenburg/loggedfs-python

	src/loggedfs/_core/cli.py: Command line interface

	Copyright (C) 2017-2019 Sebastian M. Ernst <ernst@pleiszenburg.de>

<LICENSE_BLOCK>
The contents of this file are subject to the Apache License
Version 2 ("License"). You may not use this file except in
compliance with the License. You may obtain a copy of the License at
https://www.apache.org/licenses/LICENSE-2.0
https://github.com/pleiszenburg/loggedfs-python/blob/master/LICENSE

Software distributed under the License is distributed on an "AS IS" basis,
WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License for the
specific language governing rights and limitations under the License.
</LICENSE_BLOCK>

"""


# +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
# IMPORT
# +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

import click

from .defaults import LOG_ENABLED_DEFAULT, LOG_PRINTPROCESSNAME_DEFAULT
from .fs import loggedfs_factory
from .filter import filter_pipeline_class


# +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
# ROUTINE
# +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

@click.command()
@click.option(
	'-f',
	is_flag = True,
	help = 'Do not start as a daemon. Write logs to stdout if no log file is specified.'
	)
@click.option(
	'-p',
	is_flag = True,
	help = 'Allow every user to see the new loggedfs.'
	)
@click.option(
	'-c',
	type = click.File(mode = 'r'),
	help = 'Use the "config-file" to filter what you want to log.'
	)
@click.option(
	'-s',
	is_flag = True,
	help = 'Deactivate logging to syslog.'
	)
@click.option(
	'-l',
	type = click.Path(file_okay = True, dir_okay = False, resolve_path = True),
	help = ('Use the "log-file" to write logs to.')
	)
@click.option(
	'-j', '--json',
	is_flag = True,
	help = 'Format output as JSON instead of traditional loggedfs format.'
	)
@click.option(
	'-b', '--buffers',
	is_flag = True,
	help = 'Include read/write-buffers (compressed, BASE64) in log.'
	)
@click.option(
	'--lib',
	is_flag = True,
	help = 'Run in library mode. DO NOT USE THIS FROM THE COMMAND LINE!',
	hidden = True
	)
@click.argument(
	'directory',
	type = click.Path(exists = True, file_okay = False, dir_okay = True, resolve_path = True)
	)
def cli_entry(f, p, c, s, l, json, buffers, lib, directory):
	"""LoggedFS-python is a transparent fuse-filesystem which allows to log
	every operation that happens in the backend filesystem. Logs can be written
	to syslog, to a file, or to the standard output. LoggedFS-python allows to specify an XML
	configuration file in which you can choose exactly what you want to log and
	what you don't want to log. You can add filters on users, operations (open,
	read, write, chown, chmod, etc.), filenames, commands and return code.
	"""

	loggedfs_factory(
		directory,
		**__process_config__(c, l, s, f, p, json, buffers, lib)
		)


def __process_config__(
	config_fh,
	log_file,
	log_syslog_off,
	fuse_foreground,
	fuse_allowother,
	log_json,
	log_buffers,
	lib_mode
	):

	if config_fh is not None:
		config_data = config_fh.read()
		config_fh.close()
		(
			log_enabled, log_printprocessname, filter_obj
			) = filter_pipeline_class.from_xmlstring(config_data)
		config_file = config_fh.name
	else:
		log_enabled = LOG_ENABLED_DEFAULT
		log_printprocessname = LOG_PRINTPROCESSNAME_DEFAULT
		filter_obj = filter_pipeline_class()
		config_file = None

	return {
		'fuse_foreground': fuse_foreground,
		'fuse_allowother': fuse_allowother,
		'lib_mode': lib_mode,
		'log_buffers': log_buffers,
		'_log_configfile' : config_file,
		'log_enabled': log_enabled,
		'log_file': log_file,
		'log_filter': filter_obj,
		'log_json': log_json,
		'log_printprocessname': log_printprocessname,
		'log_syslog': not log_syslog_off
		}



# -*- coding: utf-8 -*-

"""

LoggedFS-python
Filesystem monitoring with Fuse and Python
https://github.com/pleiszenburg/loggedfs-python

	src/loggedfs/_core/log.py: Logging

	Copyright (C) 2017-2019 Sebastian M. Ernst <ernst@pleiszenburg.de>

<LICENSE_BLOCK>
The contents of this file are subject to the Apache License
Version 2 ("License"). You may not use this file except in
compliance with the License. You may obtain a copy of the License at
https://www.apache.org/licenses/LICENSE-2.0
https://github.com/pleiszenburg/loggedfs-python/blob/master/LICENSE

Software distributed under the License is distributed on an "AS IS" basis,
WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License for the
specific language governing rights and limitations under the License.
</LICENSE_BLOCK>

"""


# +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
# IMPORT
# +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

import json
import logging
import logging.handlers
import os
import platform

from .timing import time


# +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
# CONST
# +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

SYSLOG_ADDRESS = {
	'Linux': '/dev/log',
	'Darwin': '/var/run/syslog'
	}


# +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
# LOGGING: Support nano-second timestamps
# +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

class _LogRecord_ns_(logging.LogRecord):

	def __init__(self, *args, **kwargs):

		self.created_ns = time.time_ns() # Fetch precise timestamp
		super().__init__(*args, **kwargs)


class _Formatter_ns_(logging.Formatter):

	default_nsec_format = '%s,%09d'

	def formatTime(self, record, datefmt=None):

		if datefmt is not None: # Do not handle custom formats here ...
			return super().formatTime(record, datefmt) # ... leave to original implementation
		ct = self.converter(record.created_ns / 1e9)
		t = time.strftime(self.default_time_format, ct)
		s = self.default_nsec_format % (t, record.created_ns - (record.created_ns // 10**9) * 10**9)
		return s


logging.setLogRecordFactory(_LogRecord_ns_)


# +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
# ROUTINES
# +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++


def get_logger(name, log_enabled, log_file, log_syslog, log_json):

	if log_json:
		log_formater = _Formatter_ns_('{"time": "%(asctime)s", "logger": "%(name)s", %(message)s}')
		log_formater_short = _Formatter_ns_('{%(message)s}')
	else:
		log_formater = _Formatter_ns_('%(asctime)s (%(name)s) %(message)s')
		log_formater_short = _Formatter_ns_('%(message)s')

	logger = logging.getLogger(name)

	if not bool(log_enabled):
		logger.setLevel(logging.CRITICAL)
		return logger
	logger.setLevel(logging.DEBUG)

	ch = logging.StreamHandler()
	ch.setLevel(logging.DEBUG)
	ch.setFormatter(log_formater)
	logger.addHandler(ch)

	if bool(log_syslog):
		try:
			sl = logging.handlers.SysLogHandler(address = SYSLOG_ADDRESS[platform.system()])
		except KeyError:
			raise NotImplementedError('unsupported operating system')
		sl.setLevel(logging.DEBUG)
		sl.setFormatter(log_formater_short)
		logger.addHandler(sl)

	if log_file is None:
		return logger

	fh = logging.FileHandler(os.path.join(log_file)) # TODO
	fh.setLevel(logging.DEBUG)
	fh.setFormatter(log_formater)
	logger.addHandler(fh)

	return logger


def log_msg(log_json, msg):

	if log_json:
		return '"msg": %s' % json.dumps(msg)
	else:
		return msg



# -*- coding: utf-8 -*-

"""

LoggedFS-python
Filesystem monitoring with Fuse and Python
https://github.com/pleiszenburg/loggedfs-python

	setup.py: Used for package distribution

	Copyright (C) 2017-2019 Sebastian M. Ernst <ernst@pleiszenburg.de>

<LICENSE_BLOCK>
The contents of this file are subject to the Apache License
Version 2 ("License"). You may not use this file except in
compliance with the License. You may obtain a copy of the License at
https://www.apache.org/licenses/LICENSE-2.0
https://github.com/pleiszenburg/loggedfs-python/blob/master/LICENSE

Software distributed under the License is distributed on an "AS IS" basis,
WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License for the
specific language governing rights and limitations under the License.
</LICENSE_BLOCK>

"""


# +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
# IMPORT
# +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

import os
from setuptools import (
	find_packages,
	setup
	)
import sys


# +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
# SETUP
# +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++


# Bump version HERE!
_version_ = '0.0.3'


# List all versions of Python which are supported
confirmed_python_versions = [
	('Programming Language :: Python :: %s' % x)
	for x in '3.4 3.5 3.6 3.7'.split(' ')
	]


# Fetch readme file
with open(os.path.join(os.path.dirname(__file__), 'README.rst')) as f:
	long_description = f.read()


# Development dependencies
development_deps_list = [
	'coverage',
	'psutil',
	'pytest',
	'python-language-server',
	'PyYAML',
	'setuptools',
	'Sphinx',
	'sphinx_rtd_theme',
	'tap.py',
	'twine',
	'wheel'
	]


# Get Python interpreter version
py_gen, py_ver, *_ = sys.version_info
# Raise an error if this is running on Python 2 (legacy)
if py_gen < 3:
	raise NotImplementedError
# Handle CPython 3.4 and below extra dependency
if py_ver < 5:
	development_deps_list.append('scandir') # See https://github.com/benhoyt/scandir


setup(
	name = 'loggedfs',
	packages = find_packages('src'),
	package_dir = {'': 'src'},
	version = _version_,
	description = 'Filesystem monitoring with Fuse and Python',
	long_description = long_description,
	author = 'Sebastian M. Ernst',
	author_email = 'ernst@pleiszenburg.de',
	url = 'https://github.com/pleiszenburg/loggedfs-python',
	download_url = 'https://github.com/pleiszenburg/loggedfs-python/archive/v%s.tar.gz' % _version_,
	license = 'Apache License 2.0',
	keywords = ['filesystem', 'fuse', 'logging', 'monitoring'],
	include_package_data = True,
	install_requires = [
		'click>=7.0',
		'fusepy @ git+https://github.com/s-m-e/fusepy@master#egg=fusepy-2.0.99',
		'xmltodict'
		],
	extras_require = {'dev': development_deps_list},
	entry_points = '''
		[console_scripts]
		loggedfs = loggedfs:cli_entry
		''',
	zip_safe = False,
	classifiers = [
		'Development Status :: 4 - Beta',
		'Environment :: Console',
		'Intended Audience :: Developers',
		'Intended Audience :: Information Technology',
		'Intended Audience :: Science/Research',
		'Intended Audience :: System Administrators',
		'License :: OSI Approved :: Apache Software License',
		'Operating System :: POSIX :: Linux',
		'Operating System :: MacOS',
		'Programming Language :: Python :: 3'
		] + confirmed_python_versions + [
		'Programming Language :: Python :: 3 :: Only',
		'Programming Language :: Python :: Implementation :: CPython',
		'Topic :: Scientific/Engineering',
		'Topic :: Software Development :: Libraries',
		'Topic :: Software Development :: Testing',
		'Topic :: System :: Archiving',
		'Topic :: System :: Filesystems',
		'Topic :: System :: Logging',
		'Topic :: System :: Monitoring',
		'Topic :: System :: Systems Administration',
		'Topic :: Utilities'
		]
)



# -*- coding: utf-8 -*-

"""

LoggedFS-python
Filesystem monitoring with Fuse and Python
https://github.com/pleiszenburg/loggedfs-python

	src/loggedfs/__init__.py: Module init

	Copyright (C) 2017-2019 Sebastian M. Ernst <ernst@pleiszenburg.de>

<LICENSE_BLOCK>
The contents of this file are subject to the Apache License
Version 2 ("License"). You may not use this file except in
compliance with the License. You may obtain a copy of the License at
https://www.apache.org/licenses/LICENSE-2.0
https://github.com/pleiszenburg/loggedfs-python/blob/master/LICENSE

Software distributed under the License is distributed on an "AS IS" basis,
WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License for the
specific language governing rights and limitations under the License.
</LICENSE_BLOCK>

"""


# +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
# IMPORT
# +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

from ._core.cli import cli_entry
from ._core.filter import (
	filter_field_class,
	filter_item_class,
	filter_pipeline_class
	)
from ._core.fs import (
	_loggedfs,
	loggedfs_factory
	)
from ._core.ipc import end_of_transmission
from ._core.notify import notify_class as loggedfs_notify
from ._core.out import decode_buffer



# -*- coding: utf-8 -*-

"""

LoggedFS-python
Filesystem monitoring with Fuse and Python
https://github.com/pleiszenburg/loggedfs-python

	src/loggedfs/_core/__init__.py: Module core init

	Copyright (C) 2017-2019 Sebastian M. Ernst <ernst@pleiszenburg.de>

<LICENSE_BLOCK>
The contents of this file are subject to the Apache License
Version 2 ("License"). You may not use this file except in
compliance with the License. You may obtain a copy of the License at
https://www.apache.org/licenses/LICENSE-2.0
https://github.com/pleiszenburg/loggedfs-python/blob/master/LICENSE

Software distributed under the License is distributed on an "AS IS" basis,
WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License for the
specific language governing rights and limitations under the License.
</LICENSE_BLOCK>

"""
